/** Account specific routers for the email service
 * in expressjs
 */
import express, { Router } from "express";
import jwt from "jsonwebtoken";
import { z } from "zod";

import { signChangeEmailToken } from "@snip/auth/tokens/changeEmail";
import { signUserToken } from "@snip/auth/tokens/emailVerify";
import { config } from "@snip/config/server";
import service from "@snip/database";

import { authMiddleWare } from "@/common/auth";
import { rateLimitAggressive } from "@/common/rateLimit";
import {
    ChangeEmailEmail,
    ForgotPasswordEmail,
    VerifyAccountEmail,
} from "@/email/email";
import { sendEmail } from "@/email/send";

export const accountRouter: Router = express.Router();

accountRouter.post(
    "/passwordReset/:email",
    rateLimitAggressive,
    async (req, res) => {
        // Parse email
        const emailAddr = req.params.email;
        const emailSchema = z.string().email();
        const parsedEmail = emailSchema.safeParse(emailAddr);
        if (!parsedEmail.success) {
            res.status(400).json({
                error: "Invalid email address",
            });
            return;
        }

        //Get pw hash for reset links token
        let pw_hash;
        try {
            pw_hash = await service.user
                .getByEmail(parsedEmail.data)
                .then(async (userData) => {
                    return await service.user.getPasswordHash(userData.email);
                });
        } catch (error) {
            if (error instanceof Error && error.message === "User not found") {
                res.status(400).json({
                    error: "User not found",
                });
                return;
            } else {
                throw error;
            }
        }

        // Generate reset link
        const data = {
            time: Date.now(),
            email: emailAddr,
            type: "pw_reset",
        };
        const token = jwt.sign(data, config.EMAIL.TOKEN_SECRET + pw_hash, {
            expiresIn: "1h",
        });

        // Create email
        const email = new ForgotPasswordEmail(parsedEmail.data, token);
        try {
            await sendEmail(email);
            res.json({
                success: true,
            });
        } catch (error) {
            console.error(
                `Failed to send forgot password email to ${parsedEmail.data}`,
            );
            console.error(error);
            res.status(500).json({
                error: "Failed to send email",
                details: (error as Error).message,
            });
        }
        return;
    },
);

/** Given a token generates sets a new password
 */
accountRouter.post(
    "/updatePassword",
    rateLimitAggressive,
    express.json(),
    async (req, res) => {
        const bodySchema = z.object({
            email: z.string().email(),
            token: z.string(),
            password: z.string().optional(),
        });
        console.log(req.body);
        const parsedBody = bodySchema.safeParse(req.body);
        if (!parsedBody.success) {
            res.status(400).json({
                error: "Invalid request",
                details: parsedBody.error?.errors,
            });
            return;
        }

        //Get pw hash for reset links token
        let pw_hash;
        try {
            pw_hash = await service.user
                .getByEmail(parsedBody.data.email)
                .then(async (userData) => {
                    return await service.user.getPasswordHash(userData.email);
                });
        } catch (error) {
            if (error instanceof Error && error.message === "User not found") {
                res.status(400).json({
                    error: "User not found",
                });
                return;
            } else {
                throw error;
            }
        }

        try {
            const data = jwt.verify(
                parsedBody.data.token,
                config.EMAIL.TOKEN_SECRET + pw_hash,
            ) as { time: number; email: string; type: string };

            if (data.email !== parsedBody.data.email) {
                throw new Error("Token validation failed");
            }
            if (data.type !== "pw_reset") {
                throw new Error("Token validation failed");
            }
        } catch (_error) {
            res.status(400).json({
                error: "Invalid token",
            });
            return;
        }

        // If no password is set return valid
        if (!parsedBody.data.password) {
            res.status(200).json({
                valid: true,
            });
            return;
        }

        // Valid token
        try {
            await service.user.updatePassword(
                parsedBody.data.email,
                parsedBody.data.password,
            );
        } catch (error) {
            res.status(500).json({
                error: "Failed to reset password",
                details: (error as Error).message,
            });
            return;
        }
        res.status(200).json({
            success: true,
        });
    },
);

/** Send email to user to request
 * verification of their email address
 */
accountRouter.post(
    "/verifyEmail",
    authMiddleWare,
    rateLimitAggressive,
    async (req, res) => {
        // Check if user exists
        let user;
        try {
            user = await service.user.getById(req.session!.user!.id);
        } catch (error) {
            if (error instanceof Error && error.message === "User not found") {
                res.status(400).json({
                    error: "User not found",
                });
                return;
            } else {
                throw error;
            }
        }

        if (user.emailValidated) {
            res.status(400).json({
                error: "Email already verified!",
            });
            return;
        }

        const token = await signUserToken(user);

        const email = new VerifyAccountEmail(user.email, token);
        try {
            await sendEmail(email);
            res.json({
                success: true,
            });
        } catch (error) {
            console.error(`Failed to send verify email to ${user.email}`);
            console.error(error);
            res.status(500).json({
                error: "Failed to send email",
                details: (error as Error).message,
            });
        }
    },
);

/** Change the email addr
 * of the user
 */
accountRouter.post(
    "/changeEmail",
    authMiddleWare,
    rateLimitAggressive,
    express.json(),
    async (req, res) => {
        const bodySchema = z.object({
            email: z.string().email(),
            password: z.string(),
        });

        const parsedBody = bodySchema.safeParse(req.body);

        if (!parsedBody.success) {
            res.status(400).json({
                error: "Invalid request body",
                details: "Valid Email and password required",
            });
            return;
        }
        const newEmail = parsedBody.data.email;
        const password = parsedBody.data.password;
        const user_id = req.session!.user!.id;

        // Check if password matches
        let valid;
        try {
            valid = await service.user.loginById(user_id, password);
        } catch (error) {
            console.error(error);
            valid = false;
        }

        if (!valid) {
            res.status(400).json({
                error: "Invalid password",
            });
            return;
        }

        // Check if email already exists
        let emailAlreadyUsed = false;
        try {
            const user = await service.user.getByEmail(newEmail);
            if (user) {
                emailAlreadyUsed = true;
            }
        } catch (error) {
            if (error instanceof Error && error.message === "User not found") {
                // Do nothing
            } else {
                res.status(500).json({
                    error: "Internal server error",
                    details: (error as Error).message,
                });
                return;
            }
        }

        if (emailAlreadyUsed) {
            res.status(400).json({
                error: "Email already registered on other account!",
            });
            return;
        }

        const user = await service.user.getById(user_id);
        const token = await signChangeEmailToken(user, parsedBody.data.email);

        const email = new ChangeEmailEmail(newEmail, token);
        try {
            await sendEmail(email);
            res.json({
                success: true,
            });
        } catch (error) {
            res.status(500).json({
                error: "Failed to send email",
                details: (error as Error).message,
            });
        }
    },
);
