"use client";
import useUser from "lib/hooks/useUser";
import { useState } from "react";
import { Button, Modal } from "react-bootstrap";
import { useFormState } from "react-dom";

import { FormError } from "components/form/alert";
import { SubmitButton } from "components/form/buttons";
import { EmailInput, PasswordInput } from "components/form/inputs/text";
import { Item } from "components/form/items";
import { validateEmail, validatePW } from "components/form/validation";

import styles from "./account.module.scss";

// unset | valid | invalid
type valid = -1 | 1 | 0;

interface ChangeEmailState {
    emailSend?: boolean;
    error?: string;
    details?: string;
    email: valid;
    confirmEmail: valid;
    password: valid;
}

const initialStateEmail: ChangeEmailState = {
    email: -1,
    password: -1,
    confirmEmail: -1,
};

export function ChangeEmail() {
    const { user, isLoading } = useUser();

    const [show, setShow] = useState(false);
    const handleClose = () => setShow(false);
    const handleShow = () => setShow(true);

    return (
        <>
            <div className="d-flex justify-content-between placeholder-glow">
                {isLoading || !user ? (
                    <>
                        <code className="placeholder col-3" />
                        <span className="placeholder col-1 " />
                    </>
                ) : (
                    <>
                        <code>{user.email}</code>
                        <span
                            className={styles.verified}
                            data-verified={user.emailValidated}
                        />
                    </>
                )}
            </div>
            <Button variant="outline-dark" onClick={handleShow}>
                Change email
            </Button>

            <Modal
                show={show}
                onHide={handleClose}
                centered
                size="lg"
                fullscreen="md-down"
                backdrop="static"
                keyboard={false}
            >
                <Modal.Header closeButton>
                    <Modal.Title>Change your email address</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <Item
                        title="New email address"
                        description="Please enter your new email address and confirm it.
                            You will receive an email on the new address to confirm the change."
                    >
                        <ChangeEmailForm />
                    </Item>
                </Modal.Body>
            </Modal>
        </>
    );
}

export function ChangeEmailForm() {
    const [state, formAction] = useFormState(changeEmail, initialStateEmail);

    return (
        <form action={formAction} className="mt-3 d-flex gap-2 flex-column">
            <FormError>
                {state.error ? (
                    <>
                        {state.error}
                        {state.details && (
                            <div className="mt-2">{state.details}</div>
                        )}
                    </>
                ) : null}
            </FormError>
            <EmailInput
                name="email"
                label="New email address"
                isValid={state.email == 1}
                isInvalid={state.email == 0}
                disabled={state.emailSend}
            />
            <EmailInput
                name="confirmEmail"
                label="Confirm new email address"
                isValid={state.confirmEmail == 1}
                isInvalid={state.confirmEmail == 0}
                invalidFeedback="Emails do not match!"
                disabled={state.emailSend}
            />
            <PasswordInput
                isValid={state.password == 1}
                isInvalid={state.password == 0}
            />
            <SubmitButton disabled={state.emailSend}>
                {state.emailSend
                    ? "Email send, check your inbox!"
                    : "Change email"}
            </SubmitButton>
        </form>
    );
}

async function changeEmail(
    _previousState: ChangeEmailState,
    formData: FormData,
): Promise<ChangeEmailState> {
    const ret: ChangeEmailState = {
        email: -1,
        password: -1,
        confirmEmail: -1,
    };
    if (formData.has("email")) {
        const email: string = formData.get("email")! as string;
        ret["email"] = validateEmail(email) ? 1 : 0;
        if (formData.has("confirmEmail")) {
            ret["confirmEmail"] =
                formData.get("email") === formData.get("confirmEmail") ? 1 : 0;
        }
    }

    if (formData.has("password")) {
        const password: string = formData.get("password")! as string;
        ret["password"] = validatePW(password) ? 1 : 0;
    }

    //Early return if any validation failed
    if (ret["email"] != 1 || ret["confirmEmail"] != 1 || ret["password"] != 1) {
        return ret;
    }

    const res = await fetch("/email/account/changeEmail", {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify({
            email: formData.get("email"),
            password: formData.get("password"),
        }),
    });

    if (!res.ok) {
        try {
            const data = await res.json();
            return {
                ...ret,
                ...data,
            };
        } catch (e) {
            console.error(e);
            return {
                ...ret,
                error: "An unknown error occurred, please try again later!",
            };
        }
    }
    const data = await res.json();

    return {
        ...ret,
        emailSend: data.success,
        error: undefined,
        details: undefined,
    };
}
