import * as t from "./sql.typings";

/** !!README!!
 * ----------------
 * All types are from an insert perspective, meaning that all but
 * the primary key fields are optional.
 *
 * If you need a full type, you can use the `Required<T>` type.
 */

/** We overload some types from the automatically generated sql typing to
 * remove unneeded fields and add the id type.
 */
export type ID = number;

export type PageData = t.pagesViewData & { id: ID };
export type BookData = t.booksData & { id: ID };
export type UserConfigData = Omit<t.user_config_resolvedData, "pen_colors"> & {
    id?: ID;
    pen_colors?: string[];
};
export type BlobData = t.blobsData & { id: ID };
export type FolderData = t.browserFoldersData & { id?: ID };
// Same as book but joins the owner and the background type table
export type BookDataResolved = t.books_resolvedData & { id: ID };
export type BackgroundTypeData = t.background_typesData & { id: ID };

type FileID = string;
/** FileData can represent both a book or a directory.
 * This is an abstraction for the book and directory data.
 */
export interface FileData {
    id: FileID; // Unique ID, database key
    name: string; // Name of the file or directory
    comment?: string; // Comment for the file

    // Properties for files
    isDir?: boolean; // If true, this is a directory, default: false
    parentId?: FileID; // ID of parent directory
    childrenIds?: FileID[]; // IDs of children

    // Properties derived from the books
    numPages?: number; // Shows number of pages in the book
    thumbnail?: string; // Automatically load thumbnail from this URL
    created?: Date; // Date the book was created
    last_updated?: Date; // Date the book was last updated
    finished?: boolean; // If true, the book is finished, default: false
    ownerID?: ID; // ID of the owner of the book
    ownerName?: string; // Username/email of the owner of the book
    url?: string; // URL to the book
    perms?: PermissionACLData; // Permissions for the book
}

// The following types are used to define the Access Control List (ACL)
export const RESOURCE_BOOK = 1;
export const RESOURCE_INSTRUMENT = 2;
export type Resource = typeof RESOURCE_BOOK | typeof RESOURCE_INSTRUMENT;
export const ENTITY_USER = 1;
export const ENTITY_GROUP = 2;
export const ENTITY_TOKEN = -10;
export type Entity =
    | typeof ENTITY_USER
    | typeof ENTITY_GROUP
    | typeof ENTITY_TOKEN;
export type PermissionACLData = t.aclData & {
    id: ID;
    entity_type: Entity;
    resource_type: Resource;
};

// TODO:create database view for bookowner
export interface BookOwner {
    id: number;
    entity_name: string;
    entity_type: "user" | "group";
}

/* -------------------------------------------------------------------------- */
/*                        snips and their serialized data                       */
/* -------------------------------------------------------------------------- */

// Snip data base type for the database services also extended
// by the different snips
export interface BaseData {}
export interface BaseView {
    x?: number | undefined;
    y?: number | undefined;
    rot?: number | undefined;
    mirror?: boolean | undefined;
}
export interface SnipData<D extends BaseData, V extends BaseView>
    extends Omit<t.snipsData, "data_json" | "view_json"> {
    data: D;
    view: V;
    type: string;
    created?: Date | undefined;
    created_by?: number | undefined | null;
}
export interface OptionalData {
    blob?: BlobData;
    snips?: SnipData<BaseData, BaseView>[];
    snip?: SnipData<BaseData, BaseView>;
}

export const BackgroundsTypes = t.background_types;

export type PartialBy<T, K extends keyof T> = Omit<T, K> & Partial<Pick<T, K>>;

export interface PermissionACLDataResolved
    extends Omit<PermissionACLData, "entity_type"> {
    entity_name: string;
    entity_type: "user" | "group";
}

/* -------------------------------------------------------------------------- */
/*                           users, groups & members                          */
/* -------------------------------------------------------------------------- */

// Just to be use remove the hashedPassword from the user data
export type UserData = Omit<t.usersData, "hashedPassword"> & { id: ID };
export type GroupData = t.groupsData & { id?: ID };
export type GroupDataWithMembers = GroupData & {
    members: (GroupMemberData & { self?: boolean })[];
};

// Did not want to create a new table for roles thus they are just
// hardcoded here for now
export type MemberRole = "invited" | "member" | "moderator" | "owner";
export type GroupMemberData = Omit<t.users_groups_with_emailData, "role"> & {
    role: MemberRole;
};

//Collaboration invites
export type CollaborationInviteData = t.collaboration_invitesData;
export type CollaborationInviteDataUsersResolved =
    t.collaboration_invites_users_resolvedData;

/* -------------------------------------------------------------------------- */
/*                          tokens and sharable links                         */
/* -------------------------------------------------------------------------- */
export type TokenData = t.tokensData;
export type APITokenData = TokenData & {
    type: "api";
};
export type UITokenData = TokenData & {
    type: "ui";
};
